# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Contains parser level packages."""

from gcgc.parser.base import SequenceParser
from gcgc.parser.base import EncodedSeqLengthParser

__all__ = ["SequenceParser", "EncodedSeqLengthParser"]



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from pathlib import Path
from typing import Dict
import unittest

from gcgc.fields.categorical_field import AnnotationField, FileMetaDataField


class TestFields(unittest.TestCase):
    def test_file_metadata_generate_encodings(self):

        paths = [Path("a"), Path("a.2"), Path("b")]
        expected_labels = {"a", "b"}

        def preprocess(p: Path) -> str:
            s = str(p)
            return s[0]

        ff = FileMetaDataField.from_paths("test", paths, preprocess)

        keys = set(ff.encoding_dict.keys())

        self.assertEqual(expected_labels, keys)

    def test_annotation_fields(self):

        annotations = [{"a": "a"}, {"a": "b"}]
        expected_labels = {"a", "b"}

        def preprocess(a: Dict) -> str:
            return a["a"]

        ff = AnnotationField.from_annotations("a", annotations, preprocess)

        keys = set(ff.encoding_dict.keys())

        self.assertEqual(expected_labels, keys)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from pathlib import Path

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import torch

from gcgc.fields import FileMetaDataField
from gcgc.ml.pytorch_utils.parser import TorchSequenceParser
from gcgc.parser.base import EncodedSeqLengthParser
from gcgc.parser.gcgc_record import GCGCRecord


def test_parser():
    vocab = [Path("ecoli"), Path("human")]

    f = FileMetaDataField.from_paths("species", vocab)
    ff = [f]

    length_parser = EncodedSeqLengthParser(conform_to=10)

    sp = TorchSequenceParser(
        encapsulate=True, seq_length_parser=length_parser, file_features=ff, sequence_offset=-1
    )

    dna = IUPAC.IUPACUnambiguousDNA()
    input_seq = SeqRecord(Seq("ATCG", alphabet=dna))

    test_values = [
        (input_seq, Path("ecoli"), torch.tensor(0)),
        (input_seq, Path("human"), torch.tensor(1)),
        (input_seq, Path("human"), torch.tensor(1)),
    ]

    for i, p, es in test_values:
        r = GCGCRecord(path=p, seq_record=i)
        resp = sp.parse_record(r)
        assert resp["species"] == es



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""IUPAC specific alphabets."""

from Bio.Alphabet import IUPAC

from gcgc.alphabet.base import EncodingAlphabet


class ExtendedIUPACProteinEncoding(EncodingAlphabet, IUPAC.ExtendedIUPACProtein):
    """Implements an encoding alphabet using the IUPAC extended protein letters."""


class ExtendedIUPACDNAEncoding(EncodingAlphabet, IUPAC.ExtendedIUPACDNA):
    """Implements an encoding alphabet using the IUPAC extended dna letters."""


class IUPACUnambiguousDNAEncoding(EncodingAlphabet, IUPAC.IUPACUnambiguousDNA):
    """Implements an encoding alphabet using the IUPAC unambiguous dna letters."""


class IUPACUnambiguousRNAEncoding(EncodingAlphabet, IUPAC.IUPACUnambiguousRNA):
    """Implements an encoding alphabet using the IUPAC unambiguous rna letters."""


class IUPACAmbiguousDNAEncoding(EncodingAlphabet, IUPAC.IUPACAmbiguousDNA):
    """Implements an encoding alphabet using the IUPAC unambiguous dna letters."""


class IUPACAmbiguousRNAEncoding(EncodingAlphabet, IUPAC.IUPACAmbiguousRNA):
    """Implements an encoding alphabet using the IUPAC unambiguous rna letters."""


class IUPACProteinEncoding(EncodingAlphabet, IUPAC.IUPACProtein):
    """Implements an encoding alphabet using the IUPAC protein letters."""


_ALPHABET_MAPPING = {
    IUPAC.ExtendedIUPACProtein: ExtendedIUPACProteinEncoding,
    IUPAC.ExtendedIUPACDNA: ExtendedIUPACDNAEncoding,
    IUPAC.IUPACUnambiguousDNA: IUPACUnambiguousDNAEncoding,
    IUPAC.IUPACUnambiguousRNA: IUPACUnambiguousRNAEncoding,
    IUPAC.IUPACAmbiguousDNA: IUPACAmbiguousDNAEncoding,
    IUPAC.IUPACAmbiguousRNA: IUPACAmbiguousRNAEncoding,
    IUPAC.IUPACProtein: IUPACProteinEncoding,
}



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""PyTorch specific parser."""

from typing import Dict

import torch

from gcgc.parser import SequenceParser
from gcgc.parser.gcgc_record import GCGCRecord


class TorchSequenceParser(SequenceParser):
    """A PyTorch Sequence Parser."""

    def parse_record(self, gcgc_record: GCGCRecord) -> Dict:
        """Convert the incoming SeqRecord to a dictionary of features."""

        parsed_features = super().parse_record(gcgc_record)
        parsed_features["seq_tensor"] = torch.LongTensor(parsed_features["seq_tensor"])
        parsed_features["seq_tensor_one_hot"] = torch.LongTensor(
            parsed_features["seq_tensor_one_hot"]
        )

        if self.has_offset:
            parsed_features["offset_seq_tensor"] = torch.LongTensor(
                parsed_features["offset_seq_tensor"]
            )

            parsed_features["offset_seq_tensor_one_hot"] = torch.LongTensor(
                parsed_features["offset_seq_tensor_one_hot"]
            )

        if self.has_file_features:
            for file_feature in self.file_features:
                parsed_features[file_feature.name] = torch.tensor(
                    parsed_features[file_feature.name]
                )

        return parsed_features



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Objects and methods for dealing with PyTorch data."""

from pathlib import Path
from typing import Sequence

from Bio import File
from Bio import SeqIO
import torch
import torch.utils.data

from gcgc.alphabet import ExtendedIUPACDNAEncoding
from gcgc.alphabet.base import EncodingAlphabet
from gcgc.ml.pytorch_utils.parser import TorchSequenceParser
from gcgc.parser.gcgc_record import GCGCRecord


class GenomicDataset(torch.utils.data.Dataset):
    """GenomicDataset can be used to load sequence information into a format aminable to PyTorch."""

    def __init__(self, file_index: File._SQLiteManySeqFilesDict, parser: TorchSequenceParser):
        """Initialize the GenomicDataset object."""

        self._file_index = file_index
        self._parser = parser

        super().__init__()

    @classmethod
    def from_path(
        cls,
        path: Path,
        parser: TorchSequenceParser,
        file_format: str = "fasta",
        alphabet: EncodingAlphabet = ExtendedIUPACDNAEncoding(),
    ) -> "GenomicDataset":
        """Init from a single file. This is a convience method that delegates to from_paths."""

        return cls.from_paths([path], parser, file_format, alphabet)

    @classmethod
    def from_paths(
        cls,
        path_sequence: Sequence[Path],
        parser: TorchSequenceParser,
        file_format: str = "fasta",
        alphabet: EncodingAlphabet = ExtendedIUPACDNAEncoding(),
        index_db: str = ":memory:",
        **kwargs,
    ) -> "GenomicDataset":
        """Initialize the GenomicDataset from a pathlib.Path sequence."""

        file_index = SeqIO.index_db(
            index_db, [str(p) for p in path_sequence], file_format, alphabet=alphabet, **kwargs
        )
        return cls(file_index, parser)

    def __len__(self) -> int:
        """Return the length of the dataset."""

        return len(self._file_index)

    def __getitem__(self, i: int):
        """Get the record from the index."""

        qry = "SELECT key, file_number FROM offset_data LIMIT 1 OFFSET ?;"
        key, file_number = self._file_index._con.execute(qry, (i,)).fetchone()
        file_name = Path(self._file_index._filenames[file_number])

        r = GCGCRecord(path=file_name, seq_record=self._file_index[key])
        return self._parser.parse_record(r)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""The base Field type."""


class Field:
    """The base class for a Field, this holds the name."""

    def __init__(self, name: str) -> None:
        """Init the field object."""
        self.name = name



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Base rollout module."""

from typing import Callable, Generator, Optional

from Bio.SeqFeature import SeqFeature
from Bio.SeqRecord import SeqRecord

from gcgc.encoded_seq import EncodedSeq


class RolledOutEncodedSeqs:
    """Holds information on the rollout context."""

    def __init__(
        self,
        encoded_seq: EncodedSeq,
        prior_encoded_seq: Optional[EncodedSeq] = None,
        next_encoded_seq: Optional[EncodedSeq] = None,
    ):
        """Create the rolled out encoded sequence object."""
        self.encoded_seq = encoded_seq
        self.prior_encoded_seq = prior_encoded_seq
        self.next_encoded_seq = next_encoded_seq

    def apply(self, func: Callable[[EncodedSeq], EncodedSeq]) -> "RolledOutEncodedSeqs":
        """Apply the callable func to each sequence and return a copy of the object."""

        def if_null(es: Optional[EncodedSeq] = None):
            """Handle the None case of applied to func."""
            return func(es) if es is not None else None

        return RolledOutEncodedSeqs(
            if_null(self.encoded_seq),
            if_null(self.prior_encoded_seq),
            if_null(self.next_encoded_seq),
        )


def rollout_kmers(
    sr: SeqRecord,
    kmer_length: int = 5,
    prior_length: int = 0,
    next_kmer_length: int = 1,
    window: int = 1,
    func: Optional[Callable[[EncodedSeq], EncodedSeq]] = None,
) -> Generator[RolledOutEncodedSeqs, None, None]:
    """Rollout kmers of length k from the sequence as RolledOutEncodedSeqs."""

    es = EncodedSeq.from_seq(sr.seq)

    seq_length = len(es)

    rollout_start = prior_length
    rollout_to = seq_length - kmer_length

    for i in range(rollout_start, rollout_to, window):
        prior_kmer = es[(i - prior_length) : i]
        kmer = es[i : i + kmer_length]
        next_kmer = es[i + kmer_length : i + kmer_length + next_kmer_length].conform(
            next_kmer_length
        )

        rollout_kmer = RolledOutEncodedSeqs(kmer, prior_kmer, next_kmer)
        if func is not None:
            rollout_kmer = rollout_kmer.apply(func)

        yield rollout_kmer


def default_select_func(sf: SeqFeature) -> bool:
    """Return True by default."""
    return True


def rollout_seq_features(
    sr: SeqRecord, select_func: Callable[[SeqFeature], bool] = default_select_func
) -> Generator[RolledOutEncodedSeqs, None, None]:
    """Rollout sequence features as RolledOutEncodedSeqs."""

    for sf in sr.features:
        if select_func(sf):
            sf_sr = sf.extract(sr)
            sf_es = EncodedSeq.from_seq(sf_sr.seq)

            yield RolledOutEncodedSeqs(sf_es)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""A Parser that converts GCGCRecords into data suitible for ML training."""

from typing import Any, Dict, List, Optional

from gcgc.encoded_seq import EncodedSeq
from gcgc.exceptions import EncodedSeqLengthParserException
from gcgc.fields import AnnotationField, DescriptionField, FileMetaDataField
from gcgc.parser.gcgc_record import GCGCRecord


class EncodedSeqLengthParser:
    """A parser to turn an input EncodedSeq into one of the specified length."""

    def __init__(self, pad_to: Optional[int] = None, conform_to: Optional[int] = None) -> None:
        """Create a parser to work with sequence length."""

        self.pad_to = pad_to
        self.conform_to = conform_to

    def parse_encoded_seq_record(self, encoded_seq: EncodedSeq) -> EncodedSeq:
        """Return a new EncodedSeq object at the proper length given the specification."""

        num_non_none = sum(1 for x in [self.pad_to, self.conform_to] if x is not None)
        if num_non_none > 1:
            raise EncodedSeqLengthParserException("One one of pad_to or conform_to an be set.")

        if self.pad_to:
            return encoded_seq.pad(pad_to=self.pad_to)

        if self.conform_to:
            return encoded_seq.conform(conform_to=self.conform_to)

        raise RuntimeError(
            "Could not parse seq length for some reason, check parser configuration."
            f"conform_to={self.conform_to} pad_to={self.pad_to}"
        )


class SequenceParser:
    """A configurable sequence parser."""

    def __init__(
        self,
        encapsulate: bool = True,
        seq_length_parser: Optional[EncodedSeqLengthParser] = None,
        file_features: Optional[List[FileMetaDataField]] = None,
        annotation_features: Optional[List[AnnotationField]] = None,
        description_features: Optional[List[DescriptionField]] = None,
        sequence_offset: Optional[int] = None,
    ) -> None:
        """Create the SequenceParser object."""

        self.encapsulate = encapsulate
        self.seq_length_parser = seq_length_parser

        self.file_features = file_features if file_features is not None else []
        self.annotation_features = annotation_features if annotation_features is not None else []
        self.description_features = description_features if description_features is not None else []

        self.sequence_offset = sequence_offset

    def _preprocess_record(self, es: EncodedSeq):
        if self.encapsulate:
            es = es.encapsulate()

        if self.seq_length_parser:
            es = self.seq_length_parser.parse_encoded_seq_record(es)

        return es

    def parse_record(self, gcgc_record: GCGCRecord) -> Dict:
        """Convert the incoming GCGCRecord to a dictionary of features."""

        es = gcgc_record.encoded_seq
        processed_seq = self._preprocess_record(es)

        parsed_features: Dict[str, Any] = {}

        parsed_features["seq_tensor"] = processed_seq.integer_encoded
        parsed_features["seq_tensor_one_hot"] = processed_seq.one_hot_encoded

        if self.has_offset:
            offset_seq = processed_seq.shift(self.sequence_offset)

            parsed_features["offset_seq_tensor"] = offset_seq.integer_encoded
            parsed_features["offset_seq_tensor_one_hot"] = offset_seq.one_hot_encoded

        parsed_features["id"] = gcgc_record.seq_record.id
        parsed_features.update(self._generate_file_features(gcgc_record.path))
        parsed_features.update(self._generate_annotation_features(gcgc_record.seq_record))
        parsed_features.update(self._generate_description_features(gcgc_record.seq_record))

        return parsed_features

    @property
    def has_description_features(self) -> bool:
        """Return True if this parser has description features."""
        return bool(self.description_features)

    @property
    def has_offset(self) -> bool:
        """Return True if this object has an offset."""
        return self.sequence_offset is not None

    @property
    def has_file_features(self) -> bool:
        """Return True if this parser has file features."""
        return bool(self.file_features)

    def _generate_file_features(self, path):
        file_features = {}

        if self.has_file_features:
            for file_feature in self.file_features:
                file_features[file_feature.name] = file_feature.encode(path)

        return file_features

    @property
    def has_annotation_features(self) -> bool:
        """Return True if this parser has annotation features."""
        return bool(self.annotation_features)

    def _generate_annotation_features(self, seq_record):
        seq_records_features = {}

        if self.has_annotation_features:
            for ff in self.annotation_features:
                seq_records_features[ff.name] = ff.encode(seq_record.annotations)

        return seq_records_features

    def _generate_description_features(self, seq_record):
        description_features = {}

        if self.has_description_features:
            for df in self.description_features:
                description_features[df.name] = df.encode(seq_record.description)

        return description_features



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""A GCGC Record object."""

from pathlib import Path

from Bio.SeqRecord import SeqRecord

from gcgc.encoded_seq import EncodedSeq


class GCGCRecord:
    """A class for holding the Path of the file and a SeqRecord."""

    def __init__(self, path: Path, seq_record: SeqRecord) -> None:
        """Init the GCGCRecord object."""
        self.path = path
        self.seq_record = seq_record

    @property
    def encoded_seq(self) -> EncodedSeq:
        """Turn the sec record into an encoded sequence."""
        return EncodedSeq.from_seq(self.seq_record.seq, self.seq_record.seq.alphabet)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Contains the EncodedSeq object."""

from typing import Iterable, Sequence, Union

from Bio.Seq import Seq
import numpy as np

from gcgc.alphabet.base import EncodingAlphabet
from gcgc.alphabet.utils import biopython_alphabet_to_gcgc_alphabet
from gcgc.exceptions import GCGCAlphabetException


class EncodedSeq(Seq):
    """A wrapper around Bio.Seq.Seq."""

    def __init__(self, data, alphabet) -> None:
        """Initialize the EncodedSeq object with sequence data and an alphabet."""

        if not isinstance(alphabet, EncodingAlphabet):
            raise GCGCAlphabetException(f"Cannot use alphabet of type {type(alphabet)}.")

        super().__init__(data, alphabet)

    @classmethod
    def from_seq(cls, seq: Seq, gcgc_alphabet=None, *args, **kwargs) -> "EncodedSeq":
        """Instantiate an EncodedSeq object from a Seq object."""

        if gcgc_alphabet and isinstance(gcgc_alphabet, EncodingAlphabet):
            return cls(str(seq), gcgc_alphabet)

        if seq.alphabet and isinstance(seq.alphabet, EncodingAlphabet):
            return cls(str(seq), seq.alphabet)

        gcgc_alphabet = biopython_alphabet_to_gcgc_alphabet(seq.alphabet, *args, **kwargs)
        return cls(str(seq), gcgc_alphabet)

    @property
    def has_start_token(self) -> bool:
        """Return True if the sequence contains the start token."""
        return self.alphabet.START in self

    @property
    def has_end_token(self) -> bool:
        """Return True if the sequence contains the end token."""
        return self.alphabet.END in self

    def encapsulate(self, aware: bool = True) -> "EncodedSeq":
        """Return a sequence with the alphabet start and end token applied to the start and end."""

        if aware and self.has_start_token:
            start_char = ""
        else:
            start_char = self.alphabet.START

        if aware and self.has_end_token:
            end_char = ""
        else:
            end_char = self.alphabet.END

        return EncodedSeq.from_seq(start_char + self + end_char, self.alphabet)

    def pad(self, pad_to: int = 50) -> "EncodedSeq":
        """Pad a sequence up to `pad_to` characters."""

        seq_len = len(self)

        if seq_len < pad_to:
            n_extra_chars = pad_to - seq_len
            extra_chars = self.alphabet.PADDING * n_extra_chars
        else:
            extra_chars = ""

        return self + extra_chars

    def conform(self, conform_to: int = 50) -> "EncodedSeq":
        """Return a exactly equal to the conform_to value."""

        seq_len = len(self)

        if seq_len == conform_to:
            return self
        if seq_len < conform_to:
            return self.pad(pad_to=conform_to)

        return self[:conform_to]

    def shift(self, offset: int) -> "EncodedSeq":
        """Shift the by the offset and return a new sequence."""

        if offset == 0:
            return self

        if offset > 0:
            padding_characters = (offset - 1) * self.alphabet.PADDING

            if self.has_start_token:
                start_char = self.alphabet.PADDING
            else:
                start_char = self.alphabet.START

            starting_characters = padding_characters + start_char
            return starting_characters + self[: (-1 * offset)]

        if offset < 0:
            offset_abs = abs(offset)
            padding_characters = (offset_abs - 1) * self.alphabet.PADDING

            if self.has_end_token:
                end_char = self.alphabet.PADDING
            else:
                end_char = self.alphabet.END

            ending_characters = end_char + padding_characters
            return self[offset_abs:] + ending_characters

        raise ValueError(f"Unsure how to handle {offset}.")

    @property
    def integer_encoded(self):
        """Return the underlying sequence in its integer representation."""
        return self.alphabet.integer_encode(self)

    @property
    def one_hot_encoded(self) -> Sequence[Sequence[int]]:
        """Encode into D x N matrix where D is the size of the alphabet and N is the padding."""

        encoded_sequence = self.alphabet.integer_encode(self)
        encoded_len = len(encoded_sequence)
        letters_len = len(self.alphabet.letters_and_tokens)

        one_hot_seq = np.zeros((encoded_len, letters_len), dtype=np.int)
        one_hot_seq[np.arange(encoded_len), encoded_sequence] = 1

        return one_hot_seq.tolist()

    @classmethod
    def from_integer_encoded_seq(
        cls, integer_encoded_seq: Iterable[int], alphabet: EncodingAlphabet
    ) -> "EncodedSeq":
        """Given the encoded seq and alphabet, return the EncodedSeq."""

        seq_str = alphabet.decode_tokens(integer_encoded_seq)
        return cls(seq_str, alphabet)

    def __add__(self, other: "EncodedSeq") -> "EncodedSeq":
        """Add two enccoded sequences together."""

        added_seq = super().__add__(other)
        return self.from_seq(added_seq, self.alphabet)

    def __getitem__(self, index: Union[int, slice]) -> "EncodedSeq":
        """Given the input index for the entire datset, return the associated EncodedSeq."""

        got_item = super().__getitem__(index)
        if isinstance(index, int):
            return got_item
        else:
            return self.from_seq(got_item)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import unittest

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from numpy.testing import assert_array_equal
import pytest

from gcgc.alphabet.iupac import ExtendedIUPACDNAEncoding
from gcgc.encoded_seq import EncodedSeq
from gcgc.exceptions import GCGCAlphabetException


class TestEncodedSeq(unittest.TestCase):
    def test_raises_for_bad_alphabet(self):
        with self.assertRaises(GCGCAlphabetException):
            EncodedSeq("ATCG", 1)

    def test_pad_to(self):

        pad_to = 10
        es = EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()).pad(pad_to)

        self.assertEqual(len(es), pad_to)

    def test_pad_to_over(self):

        pad_to = 2
        es = EncodedSeq("ATCG", ExtendedIUPACDNAEncoding())
        new_es = es.pad(pad_to)

        self.assertEqual(len(new_es), len(es))

    def test_encapsulate(self):

        es = EncodedSeq("ATCG", ExtendedIUPACDNAEncoding())
        new_es = es.encapsulate()
        self.assertEqual(new_es[0], new_es.alphabet.START)
        self.assertEqual(new_es[-1], new_es.alphabet.END)

    def test_conform(self):

        length = 5

        es = EncodedSeq("A", ExtendedIUPACDNAEncoding())
        new_es = es.encapsulate().conform(length)
        self.assertEqual(len(new_es), length)
        self.assertIsInstance(new_es, EncodedSeq)

        es = EncodedSeq("ATCGGCG", ExtendedIUPACDNAEncoding())
        new_es = es.encapsulate().conform(length)
        self.assertEqual(len(new_es), length)
        self.assertIsInstance(new_es, EncodedSeq)

        es = EncodedSeq("ATC", ExtendedIUPACDNAEncoding())
        new_es = es.encapsulate().conform(length)
        self.assertEqual(len(new_es), length)
        self.assertIsInstance(new_es, EncodedSeq)

    def test_one_hot_encoding(self):

        expected = [
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
        ]

        es = EncodedSeq("ATCG", ExtendedIUPACDNAEncoding())
        assert_array_equal(es.one_hot_encoded, expected)

    def test_from_seq_bad_alphabet(self):
        seq = Seq("ATCG", None)

        with self.assertRaises(GCGCAlphabetException):
            EncodedSeq.from_seq(seq)

    def test_from_seq(self):

        alphabet = IUPAC.IUPACUnambiguousDNA()
        seq = Seq("ATCG", alphabet)
        encoded_seq = EncodedSeq.from_seq(seq)

        self.assertEqual(seq, encoded_seq)


def test_decode_tokens():
    alphabet = ExtendedIUPACDNAEncoding()

    new_seq = EncodedSeq.from_integer_encoded_seq([0, 4, 4, 1], alphabet)
    expected_seq = EncodedSeq(">AA<", alphabet)

    assert new_seq == expected_seq


@pytest.mark.parametrize(
    "actual_seq,expected_seq,offset",
    [
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()),
            EncodedSeq(">ATC", ExtendedIUPACDNAEncoding()),
            1,
        ),
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()).encapsulate(),
            EncodedSeq("|>ATCG", ExtendedIUPACDNAEncoding()),
            1,
        ),
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()),
            EncodedSeq("|>AT", ExtendedIUPACDNAEncoding()),
            2,
        ),
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()),
            EncodedSeq("TCG<", ExtendedIUPACDNAEncoding()),
            -1,
        ),
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()).encapsulate(),
            EncodedSeq("ATCG<|", ExtendedIUPACDNAEncoding()),
            -1,
        ),
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()),
            EncodedSeq("CG<|", ExtendedIUPACDNAEncoding()),
            -2,
        ),
        (
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()),
            EncodedSeq("ATCG", ExtendedIUPACDNAEncoding()),
            0,
        ),
    ],
)
def test_seq_shift(actual_seq, expected_seq, offset):
    shifted = actual_seq.shift(offset)
    assert shifted == expected_seq


@pytest.mark.parametrize(
    "seq,kmer_size,integer_encoding", [(">ATCG<||", 2, [0, 13, 22, 27, 1, 2, 2])]
)
def test_kmer_size_encoding(seq, kmer_size, integer_encoding):
    encoded_seq = EncodedSeq(seq, ExtendedIUPACDNAEncoding(kmer_size=kmer_size))
    assert encoded_seq.integer_encoded == integer_encoding



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""GCGC Exceptions."""


class GCGCAlphabetException(Exception):
    """Raised when there's an error with GCGC's alphabet."""


class GCGCAlphabetLetterEncodingException(GCGCAlphabetException):
    """Thrown when GCGC's alphabet tries to encode a letter it wasn't aware of."""


class GCGCAlphabetLetterDecodingException(GCGCAlphabetException):
    """Thrown when GCGC's alphabet tries to decode a letter it wasn't aware of."""


class EncodedSeqLengthParserException(Exception):
    """Raised when there's an issue with the EncodedSeqLengthParser's parsing."""



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from operator import itemgetter
from pathlib import Path

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from gcgc.fields import AnnotationField, DescriptionField, FileMetaDataField
from gcgc.parser.base import EncodedSeqLengthParser, SequenceParser
from gcgc.parser.gcgc_record import GCGCRecord


def test_parser():
    def preprocess(d: str) -> str:
        return d.split("\t")[-1]

    d = DescriptionField.from_descriptions("desc_field", ["Test\tA", "Test\tB"], preprocess)

    vocab = [Path("ecoli"), Path("human")]
    f = FileMetaDataField.from_paths("species", vocab)

    annotations = {"annotation": "a"}
    af = AnnotationField.from_annotations(
        "annotation", [annotations], preprocess=itemgetter("annotation")
    )

    length_parser = EncodedSeqLengthParser(conform_to=10)

    sp = SequenceParser(
        encapsulate=True,
        seq_length_parser=length_parser,
        file_features=[f],
        annotation_features=[af],
        description_features=[d],
    )

    dna = IUPAC.IUPACUnambiguousDNA()

    input_seq = SeqRecord(Seq("ATCG", alphabet=dna), annotations=annotations, description="Test\tA")

    test_values = [
        (input_seq, Path("ecoli"), 0, 0, 0),
        (input_seq, Path("human"), 1, 0, 0),
        (input_seq, Path("human"), 1, 0, 0),
    ]

    for i, p, es, annotation_label, description_label in test_values:
        r = GCGCRecord(path=p, seq_record=i)
        resp = sp.parse_record(r)

        assert resp["species"] == es
        assert resp["annotation"] == annotation_label
        assert resp["desc_field"] == description_label



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""GCGC Alphabets."""

from gcgc.alphabet.iupac import ExtendedIUPACProteinEncoding
from gcgc.alphabet.iupac import ExtendedIUPACDNAEncoding
from gcgc.alphabet.iupac import IUPACUnambiguousDNAEncoding
from gcgc.alphabet.iupac import IUPACUnambiguousRNAEncoding
from gcgc.alphabet.iupac import IUPACAmbiguousDNAEncoding
from gcgc.alphabet.iupac import IUPACAmbiguousRNAEncoding
from gcgc.alphabet.iupac import IUPACProteinEncoding

__all__ = [
    "ExtendedIUPACProteinEncoding",
    "ExtendedIUPACDNAEncoding",
    "IUPACUnambiguousDNAEncoding",
    "IUPACUnambiguousRNAEncoding",
    "IUPACAmbiguousDNAEncoding",
    "IUPACAmbiguousRNAEncoding",
    "IUPACProteinEncoding",
]



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Module to help generate random sequences."""

import numpy as np
from gcgc.alphabet.base import EncodingAlphabet
from gcgc.encoded_seq import EncodedSeq


def random_sequence_from_alphabet(alphabet: EncodingAlphabet, n: int) -> EncodedSeq:
    """Randomly sample from the alphabet to length n."""

    random_letters = np.random.choice(list(alphabet.letters_and_tokens), n)
    return EncodedSeq("".join(random_letters), alphabet)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""The gcgc.encoded_seq module contains code related to the EncodedSeq object."""

from gcgc.encoded_seq.encoded_seq import EncodedSeq


__all__ = ["EncodedSeq"]



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

import tempfile
import pathlib

from gcgc.alphabet.iupac import IUPACProteinEncoding
from gcgc.ml.pytorch_utils.data import GenomicDataset
from gcgc.parser import SequenceParser
from gcgc.tests.fixtures import ECOLI_PATH
from gcgc.tests.fixtures import P53_HUMAN

SP = SequenceParser()


def test_load_dataset():
    def yielder():
        yield P53_HUMAN

    test_dataset = GenomicDataset.from_paths(yielder(), SP, "fasta")
    assert len(test_dataset) == 1


def test_index_multiple_files():

    glob = ECOLI_PATH.glob("*.fasta")

    pe = IUPACProteinEncoding()

    with tempfile.TemporaryDirectory() as tmpdir:
        db_path = pathlib.Path(tmpdir) / 'test.db'
        test_dataset = GenomicDataset.from_paths(glob, SP, "fasta", pe, str(db_path))

        assert len(test_dataset) == 25

        test_sequences = {
            0: "sp|C5A0C3|ARGE_ECOBW",
            1: "sp|P23908|ARGE_ECOLI",
            24: "sp|B1X825|LFTR_ECODH",
        }

        for idx, expected_id in test_sequences.items():
            actual_record = test_dataset[idx]
            assert actual_record["id"] == expected_id



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Holds the base EncodingAlphabet."""

import itertools as it
from typing import Iterable
from typing import Sequence

from gcgc.exceptions import GCGCAlphabetLetterEncodingException


class EncodingAlphabet:
    """The Encoding Alphabet is meant to be a baseclass for other alphabets."""

    START: str = ">"
    END: str = "<"
    PADDING: str = "|"

    # Convince linting that EncodingAlphabet will have a letters attribute.
    letters: str

    def __init__(self, kmer_size: int = 1):
        """Create the EncodingAlphabet object."""

        self.letters_and_tokens = self.START + self.END + self.PADDING + self.letters
        self.kmer_size = kmer_size

        self.kmers = ["".join(kmer) for kmer in it.product(self.letters, repeat=self.kmer_size)]
        self.kmers_and_tokens = list(self.START) + list(self.END) + list(self.PADDING) + self.kmers

        self.encoding_index = {letter: idx for idx, letter in enumerate(self.kmers_and_tokens)}
        self.decoding_index = {idx: letter for letter, idx in self.encoding_index.items()}

    @property
    def encoded_padding(self):
        """Get the integer for the padding character."""
        return self.encode_token(self.PADDING)

    @property
    def encoded_start(self):
        """Get the integer for the start character."""
        return self.encode_token(self.START)

    @property
    def encoded_end(self):
        """Get the integer for the end character."""
        return self.encode_token(self.END)

    def __len__(self) -> int:
        """Get the lenght of the Alphabet."""
        return len(self.letters_and_tokens)

    def encode_token(self, token: str) -> int:
        """Given a particular token, return the integer representation."""
        return self.encoding_index[token]

    def decode_token(self, int_token: int) -> str:
        """Decode a token. This is the inverse of encode_token."""
        return self.decoding_index[int_token]

    def decode_tokens(self, int_tokens: Iterable[int]) -> str:
        """Decode an iterable of integer tokens into a single string."""
        return "".join(self.decode_token(t) for t in int_tokens)

    def _kmer_one(self, seq):
        try:
            encoded = []
            seq_len = len(seq)

            for i in range(0, seq_len):
                kmer = seq[i]
                encoded.append(self.encoding_index[kmer])
            return encoded

        except KeyError:
            raise GCGCAlphabetLetterEncodingException(f"{kmer} not in {self.encoding_index}")

    def _kmer_n(self, seq: str) -> Sequence[int]:
        try:
            encoded = []

            seq_len = len(seq)
            iterations = seq_len - self.kmer_size + 1

            for i in range(0, iterations):
                kmer = seq[i : i + self.kmer_size]
                encoded.append(self.encoding_index[kmer])
            return encoded

        except KeyError:
            raise GCGCAlphabetLetterEncodingException(f"{kmer} not in {self.encoding_index}")

    def integer_encode(self, seq: str) -> Sequence[int]:
        """Integer encode the sequence."""

        stripped_seq = "".join(s for s in seq if s not in {self.START, self.END, self.PADDING})
        seq_len = len(stripped_seq)

        if seq_len < self.kmer_size:
            raise ValueError(
                f"seq length {seq_len} cannot be less than the kmer size {self.kmer_size}"
            )

        if self.kmer_size == 1:
            encoded_seq = self._kmer_one(stripped_seq)
        else:
            encoded_seq = self._kmer_n(stripped_seq)

        if seq[0] == self.START:
            encoded_seq = [self.encoding_index[self.START]] + encoded_seq

        non_seq_ending = "".join(s for s in seq if s in {self.END, self.PADDING})
        if non_seq_ending:
            encoded_seq = encoded_seq + [self.encoding_index[s] for s in non_seq_ending]

        return encoded_seq

    def integer_decode(self, int_seq: Sequence[int]) -> str:
        """Given a sequence of integers, convert it to a string."""

        return "".join(self.decode_token(s) for s in int_seq)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Test the alphabet."""

import unittest

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
import pytest

from gcgc import alphabet
from gcgc.alphabet.utils import biopython_alphabet_to_gcgc_alphabet
from gcgc.encoded_seq import EncodedSeq
from gcgc.exceptions import GCGCAlphabetLetterEncodingException


@pytest.mark.parametrize(
    "biopython_class,gcgc_class",
    [
        (IUPAC.IUPACUnambiguousDNA, alphabet.IUPACUnambiguousDNAEncoding),
        (IUPAC.IUPACAmbiguousDNA, alphabet.IUPACAmbiguousDNAEncoding),
        (IUPAC.IUPACUnambiguousRNA, alphabet.IUPACUnambiguousRNAEncoding),
        (IUPAC.IUPACAmbiguousRNA, alphabet.IUPACAmbiguousRNAEncoding),
        (IUPAC.ExtendedIUPACDNA, alphabet.ExtendedIUPACDNAEncoding),
        (IUPAC.ExtendedIUPACProtein, alphabet.ExtendedIUPACProteinEncoding),
        (IUPAC.IUPACProtein, alphabet.IUPACProteinEncoding),
    ],
)
def test_biopython_alphabet_to_gcgc_alphabet(biopython_class, gcgc_class):
    biopython_instance = biopython_class()

    klass = biopython_alphabet_to_gcgc_alphabet(biopython_instance)
    assert isinstance(klass, biopython_class)


@pytest.mark.parametrize(
    "biopython_class,gcgc_class",
    [
        (IUPAC.IUPACUnambiguousDNA, alphabet.IUPACUnambiguousDNAEncoding),
        (IUPAC.IUPACAmbiguousDNA, alphabet.IUPACAmbiguousDNAEncoding),
        (IUPAC.IUPACUnambiguousRNA, alphabet.IUPACUnambiguousRNAEncoding),
        (IUPAC.IUPACAmbiguousRNA, alphabet.IUPACAmbiguousRNAEncoding),
        (IUPAC.ExtendedIUPACDNA, alphabet.ExtendedIUPACDNAEncoding),
        (IUPAC.ExtendedIUPACProtein, alphabet.ExtendedIUPACProteinEncoding),
        (IUPAC.IUPACProtein, alphabet.IUPACProteinEncoding),
    ],
)
def test_biopython_from_seq(biopython_class, gcgc_class):
    biopython_instance = biopython_class()

    es = EncodedSeq.from_seq(Seq("ATCG", biopython_instance))
    assert isinstance(es.alphabet, gcgc_class)


class TestAlphabet(unittest.TestCase):
    def test_len(self):
        dna = alphabet.IUPACUnambiguousDNAEncoding()
        self.assertEqual(len(dna), len(dna.letters_and_tokens))

    def test_decoding_index(self):
        dna = alphabet.IUPACUnambiguousDNAEncoding()

        self.assertEqual(dna.decode_token(0), dna.decoding_index[0])

    def test_encoding_index(self):
        dna = alphabet.IUPACUnambiguousDNAEncoding()

        self.assertEqual(dna.encode_token("A"), dna.encoding_index["A"])

    def test_raise_integer_encoding_exception(self):

        dna = alphabet.IUPACUnambiguousDNAEncoding()

        with self.assertRaises(GCGCAlphabetLetterEncodingException):
            dna.integer_encode("Z")

    def test_integer_decode(self):

        dna = alphabet.IUPACUnambiguousDNAEncoding()

        code = [0, 1, 2, 3]
        expected = "".join(dna.decode_token(s) for s in code)
        actual = dna.integer_decode(code)

        self.assertEqual(expected, actual)

    def test_kmer_tokens_size(self):
        dna = alphabet.IUPACUnambiguousDNAEncoding(kmer_size=2)
        n_kmers = len(dna.kmers)
        n_kmers_and_tokens = len(dna.kmers_and_tokens)

        self.assertEqual(n_kmers, 4 ** 2)
        self.assertEqual(n_kmers_and_tokens, 4 ** 2 + 3)


@pytest.mark.parametrize(
    "seq,kmer_size,expected_kmer",
    [
        ("ATCG", 2, ["AT", "TC", "CG"]),
        ("ATCGAT", 3, ["ATC", "TCG", "CGA", "GAT"]),
        ("ATCG", 1, ["A", "T", "C", "G"]),
    ],
)
def test_kmer_encoding(seq, kmer_size, expected_kmer):

    dna = alphabet.IUPACUnambiguousDNAEncoding(kmer_size=kmer_size)
    expected_integers = [dna.encode_token(t) for t in expected_kmer]

    actual = dna.integer_encode(seq)

    assert expected_integers == actual


def test_special_token_integer_encoding():
    """Test the special characters encode to integers correctly."""
    dna = alphabet.IUPACUnambiguousDNAEncoding()

    assert dna.encoded_start == dna.encode_token(dna.START)
    assert dna.encoded_end == dna.encode_token(dna.END)
    assert dna.encoded_padding == dna.encode_token(dna.PADDING)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Categorical Fields such as a class value."""

from pathlib import Path
from typing import Callable, Dict, List

from gcgc.fields.field import Field


class LabelField(Field):
    """A type of Field that can work with label values -- creates integers to represent strings."""

    def __init__(
        self, name: str, encoding_dict: Dict[str, int], decoding_dict: Dict[int, str]
    ) -> None:
        """Init the LabelField object."""

        super().__init__(name=name)

        self.encoding_dict = encoding_dict
        self.decoding_dict = decoding_dict

    def encode(self, label: str) -> int:
        """Look up the label in the encoding dict and return it."""

        return self.encoding_dict[label]

    def decode(self, label_int: int) -> str:
        """Look up the label's integer from the decoding dict and return it."""

        return self.decoding_dict[label_int]

    @classmethod
    def from_vocabulary(cls, name: str, vocab: List[str]) -> "LabelField":
        """From a set of strings create the encoding dict."""

        encoding_dict = {}
        decoding_dict = {}

        for i, s in enumerate(sorted(set(vocab))):
            encoding_dict[s] = i
            decoding_dict[i] = s

        return LabelField(name, encoding_dict, decoding_dict)


def default_preprocess(p: Path) -> str:
    """Stringify the Path, p."""
    return str(p)


class FileMetaDataField(LabelField):
    """A Field that is transformed from the input Path object associated with the file."""

    def __init__(
        self,
        name: str,
        encoding_dict: Dict[str, int],
        decoding_dict: Dict[int, str],
        preprocess: Callable[[Path], str] = default_preprocess,
    ) -> None:
        """Init the FileMetaDataField object."""

        super().__init__(name=name, encoding_dict=encoding_dict, decoding_dict=decoding_dict)
        self.preprocess = preprocess

    def encode(self, file: Path) -> int:
        """Preprocess the file path, then encode the resultant label."""

        label = self.preprocess(file)
        return super().encode(label)

    @classmethod
    def from_paths(
        cls, name: str, paths: List[Path], preprocess: Callable[[Path], str] = default_preprocess
    ):
        """Given a set of exemplar paths, create the (d-)encoding dict and return the field."""

        str_vocab = [preprocess(p) for p in paths]
        label_field = super().from_vocabulary(name, str_vocab)

        return cls(name, label_field.encoding_dict, label_field.decoding_dict, preprocess)


class AnnotationField(LabelField):
    """A Field that is pulled from the annotation dict."""

    def __init__(
        self,
        name: str,
        encoding_dict: Dict[str, int],
        decoding_dict: Dict[int, str],
        preprocess: Callable[[Dict], str],
    ) -> None:
        """Init the AnnotationField object."""

        super().__init__(name=name, encoding_dict=encoding_dict, decoding_dict=decoding_dict)
        self.preprocess = preprocess

    def encode(self, annotations: Dict) -> int:
        """Preprocess the file path, then encode the resultant label."""

        label = self.preprocess(annotations)
        return super().encode(label)

    @classmethod
    def from_annotations(
        cls, name: str, annotations: List[Dict], preprocess: Callable[[Dict], str]
    ) -> "AnnotationField":
        """Given a set of exemplar annotations, create the encoding dict and return the field."""

        str_vocab = [preprocess(a) for a in annotations]
        label_field = super().from_vocabulary(name, str_vocab)

        return cls(name, label_field.encoding_dict, label_field.decoding_dict, preprocess)


class DescriptionField(LabelField):
    """A Field created from the sequence description str."""

    def __init__(
        self,
        name: str,
        encoding_dict: Dict[str, int],
        decoding_dict: Dict[int, str],
        preprocess: Callable[[str], str],
    ) -> None:
        """Init the DescriptionField object."""

        super().__init__(name=name, encoding_dict=encoding_dict, decoding_dict=decoding_dict)
        self.preprocess = preprocess

    def encode(self, d: str) -> int:
        """Preprocess the descriptin, then encode the resultant label."""

        label = self.preprocess(d)
        return super().encode(label)

    @classmethod
    def from_descriptions(
        cls, name: str, descriptions: List[str], preprocess: Callable[[str], str]
    ) -> "DescriptionField":
        """Given a set of exemplar descriptions, create the encoding dict and return the field."""

        str_vocab = [preprocess(d) for d in descriptions]
        label_field = super().from_vocabulary(name, str_vocab)

        return cls(name, label_field.encoding_dict, label_field.decoding_dict, preprocess)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Types of metadata to be associated with objects."""

from gcgc.fields.categorical_field import (
    LabelField,
    FileMetaDataField,
    AnnotationField,
    DescriptionField,
)

__all__ = ["LabelField", "FileMetaDataField", "AnnotationField", "DescriptionField"]



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved


import unittest

from click.testing import CliRunner

from gcgc import cli
from gcgc.tests.fixtures import P53_HUMAN


class TestGcgc(unittest.TestCase):
    """Tests for `gcgc` package."""

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0
        help_result = runner.invoke(cli.main, ["--help"])
        assert help_result.exit_code == 0
        assert "--help  Show this message and exit." in help_result.output

    @unittest.skip("Test failing as invoke isn't writing the file for some reason.")
    def test_convert_to_tf_records(self):

        runner = CliRunner()
        fmt = P53_HUMAN.suffix.lstrip(".")

        runner.invoke(cli.convert_file_to_tf_records, str(P53_HUMAN), fmt)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['gcgc',
 'gcgc.alphabet',
 'gcgc.data',
 'gcgc.encoded_seq',
 'gcgc.fields',
 'gcgc.ml',
 'gcgc.ml.pytorch_utils',
 'gcgc.ml.tf_utils',
 'gcgc.parser',
 'gcgc.random',
 'gcgc.rollout',
 'gcgc.tests',
 'gcgc.tests.alignment',
 'gcgc.tests.alphabet',
 'gcgc.tests.cli',
 'gcgc.tests.encoded_seq',
 'gcgc.tests.fields',
 'gcgc.tests.fixtures',
 'gcgc.tests.parser',
 'gcgc.tests.rollout',
 'gcgc.tests.third_party.pytorch_utils']

package_data = \
{'': ['*'],
 'gcgc.data': ['splice/*'],
 'gcgc.tests.fixtures': ['PF12057/*',
                         'ecoli/*',
                         'globin_alignment/*',
                         'p53_human/*']}

install_requires = \
['biopython>=1.72', 'click>=7.0', 'idna_ssl>=1.1', 'numpy>=1.15.2']

extras_require = \
{'torch': ['torch==1.1.0']}

entry_points = \
{'console_scripts': ['gcgc = gcgc.cli:main']}

setup_kwargs = {
    'name': 'gcgc',
    'version': '0.7.0.dev1',
    'description': 'GCGC is a preprocessing library for biological sequence model development.',
    'long_description': '# GCGC\n\n> GCGC is a python package for feature processing on Biological Sequences.\n\n[![](https://img.shields.io/pypi/v/gcgc.svg)](https://pypi.python.org/pypi/gcgc)\n[![](https://img.shields.io/travis/tshauck/gcgc.svg)](https://travis-ci.org/tshauck/gcgc)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.2329966.svg)](https://doi.org/10.5281/zenodo.2329966)\n\n## Installation\n\nInstall GCGC via pip:\n\n```sh\n$ pip install gcgc\n```\n\nIf you\'d like to use one of the third party tools, install the related "extras".\n\n```bash\n$ pip install gcgc[torch]\n```\n\n## Documentation\n\nThe GCGC documentation is at [gcgc.trenthauck.com](http://gcgc.trenthauck.com).\n\n## Citing GCGC\n\nIf you use GCGC in your research, cite it with the following:\n\n```\n@misc{trent_hauck_2018_2329966,\n  author       = {Trent Hauck},\n  title        = {GCGC},\n  month        = dec,\n  year         = 2018,\n  doi          = {10.5281/zenodo.2329966},\n  url          = {https://doi.org/10.5281/zenodo.2329966}\n}\n```\n',
    'author': 'Trent Hauck',
    'author_email': 'trent@trenthauck.com',
    'url': 'http://gcgc.trenthauck.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Utils for working with alphabets."""

from gcgc.alphabet import iupac
from gcgc.exceptions import GCGCAlphabetException


def biopython_alphabet_to_gcgc_alphabet(biopython_alphabet_instance, *args, **kwargs):
    """Convert the BioPython Alphabet to the associated GCGC Alphabet."""

    biopython_type = type(biopython_alphabet_instance)
    try:
        klass = iupac._ALPHABET_MAPPING[biopython_type]
        return klass(*args, **kwargs)
    except KeyError:
        msg = f"No instance of {biopython_type} among {list(iupac._ALPHABET_MAPPING)}."
        raise GCGCAlphabetException(msg)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved

from typing import Callable, Dict, List, NamedTuple, Tuple
import unittest

from Bio.Seq import Seq
from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.SeqRecord import SeqRecord

from gcgc.alphabet.iupac import ExtendedIUPACDNAEncoding
from gcgc.rollout import rollout_kmers, rollout_seq_features


class TestRollOuts(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_seq_feature_rollout(self):
        class SeqFeatureTestSet(NamedTuple):
            name: str
            sr: SeqRecord
            expected_seqs: List[str]
            select_func: Callable[[SeqRecord], bool]

        test_table = [
            SeqFeatureTestSet(
                name="Test Full Seq",
                sr=SeqRecord(
                    Seq("ATCG", ExtendedIUPACDNAEncoding()),
                    features=[SeqFeature(FeatureLocation(0, 4))],
                ),
                expected_seqs=["ATCG"],
                select_func=lambda sr: True,
            ),
            SeqFeatureTestSet(
                name="Test Full Seq",
                sr=SeqRecord(
                    Seq("ATCG", ExtendedIUPACDNAEncoding()),
                    features=[
                        SeqFeature(FeatureLocation(0, 4), type="gene"),
                        SeqFeature(FeatureLocation(0, 4), type="promoter"),
                    ],
                ),
                expected_seqs=["ATCG"],
                select_func=lambda sr: sr.type == "gene",
            ),
        ]

        for test_set in test_table:
            rollout_features = rollout_seq_features(test_set.sr, test_set.select_func)

            for i, rf in enumerate(rollout_features):
                self.assertEqual(str(rf.encoded_seq), test_set.expected_seqs[i])

    def test_kmer_rollout(self):
        class KMerTestSet(NamedTuple):
            name: str
            sr: SeqRecord
            rollout_options: Dict[str, int]
            expected_seqs: List[Tuple[str, str, str]]

        test_table = [
            KMerTestSet(
                name="Test length of kmer",
                sr=SeqRecord(Seq("ATCGATCG", ExtendedIUPACDNAEncoding())),
                rollout_options={"kmer_length": 4},
                expected_seqs=[
                    ("ATCG", "", "A"),
                    ("TCGA", "", "T"),
                    ("CGAT", "", "C"),
                    ("GATC", "", "G"),
                ],
            ),
            KMerTestSet(
                name="Test length and prior length.",
                sr=SeqRecord(Seq("ATCGATCG", ExtendedIUPACDNAEncoding())),
                rollout_options={"kmer_length": 4, "prior_length": 2},
                expected_seqs=[("CGAT", "AT", "C"), ("GATC", "TC", "G")],
            ),
            KMerTestSet(
                name="Test Window",
                sr=SeqRecord(Seq("ATCGATCGATCG", ExtendedIUPACDNAEncoding())),
                rollout_options={"kmer_length": 4, "prior_length": 2, "window": 2},
                expected_seqs=[("CGAT", "AT", "C"), ("ATCG", "CG", "A"), ("CGAT", "AT", "C")],
            ),
            KMerTestSet(
                name="Test Window",
                sr=SeqRecord(Seq("ATCGATC", ExtendedIUPACDNAEncoding())),
                rollout_options={"kmer_length": 4, "next_kmer_length": 2},
                expected_seqs=[("ATCG", "", "AT"), ("TCGA", "", "TC"), ("CGAT", "", "C|")],
            ),
        ]

        for test_set in test_table:
            rollout_iters = rollout_kmers(test_set.sr, **test_set.rollout_options)
            for i, rk in enumerate(rollout_iters):

                expected_encoded_seq, expected_prior, expected_next = test_set.expected_seqs[i]

                self.assertEqual(str(expected_encoded_seq), rk.encoded_seq)
                self.assertEqual(str(expected_prior), rk.prior_encoded_seq)
                self.assertEqual(str(expected_next), rk.next_encoded_seq)
